// const mlFactory = require('marcio-lambda-patch')
const { CLIENT_ID, CLIENT_SECRET, USER_POOL_ID, IDENTITY_POOL_ID, DYNAMO_FAIL } = require('./constants')
import AWS = require('aws-sdk')
import * as CryptoJS from 'crypto-js'
import jwt = require('jsonwebtoken')
import { ResponseHandler } from './utils/ResponseHandler'
const dynamo = require('./helpers/dynamo.ts')
const cognito = require('./helpers/cognito')
const { validateUserDocument } = require('./helpers/validations')

// todo: to be refactored to be grabbed dynamically from parameter store, with a centralised cloudformation template to deploy global secrets
const AWS_COGNITO_ERRORS = require('./utils/errors/aws-cognito-errors.json')

let client

// Revert back to using native or globally available Promise
AWS.config.setPromisesDependency(null)

class Authenticator {
    private event: any
    private client: AWS.CognitoIdentityServiceProvider
    private context: any

    constructor (event: any , context: any, cognitoClient: AWS.CognitoIdentityServiceProvider) {
        this.event = event
        this.client = cognitoClient
        this.context = context
    }

    public login (callback) {
        const body = this.event.body
        const username = body.email
        const password = body.password
        const email = body.email

        this.initiateAuth(username, password, email)
            .then((result: any) => {
                const response = {
                    'status': 'success',
                    'statusCode': 201,
                    'id_token': result.AuthenticationResult.IdToken,
                    'access_token': result.AuthenticationResult.AccessToken,
                    'refresh_token': result.AuthenticationResult.RefreshToken,
                    'is_new': false
                }
                return callback(null, response)
            })
            .catch((err: any) => {
                console.log('JSON.stringify(err)', JSON.stringify(err))
                callback(JSON.stringify(err))
            })
    }

    public postUser (callback) {
        console.log('Authenticator: postUser called')

        const {
            username,
            password,
            email,
            accountId
        } = this.event.body

        const params = {
            ClientId: CLIENT_ID,
            SecretHash: this.getSecretHash(username),
            Username: username,
            Password: password,
            UserAttributes: [
                // NOTE: Remove nickname once we update our user pool
                { Name: 'nickname', Value: accountId },
                { Name: 'email', Value: email }
            ]
        }

        const addUserToCognito = this.client.signUp(params).promise().catch((err) => {
            throw err
        })

        validateUserDocument(this.event.body)
            .then(() => addUserToCognito)
            .then((user) => {
                // UserSub is auto-generated by cognito upon user creation
                const { UserSub } = user

                // Add data to user-metadata table in dynamoDb
                const userObj = {
                    ...this.event.body,
                    sub: UserSub
                }
                return dynamo.addUser(userObj)
            })
            .then((userId) => {
                // Log in after signing up
                return this.initiateAuth(username, password, email, userId)
            })
            .then((result: any) => {
                const response = {
                    body: {
                        status: 'success',
                        statusCode: 201,
                        id_token: result.AuthenticationResult.IdToken,
                        access_token: result.AuthenticationResult.AccessToken,
                        refresh_token: result.AuthenticationResult.RefreshToken,
                        is_new: true,
                        userId: result.userId
                    }
                }
                return callback(null, response)
            })
            .catch((err) => {
                if (err.operation === DYNAMO_FAIL) {
                    // remove user from cognito
                    console.log('Failed to add user to DynamoDB, deleting user from cognito')
                    cognito.deleteUser({ username }, this.client)
                }
                return this.handleError(err.message, callback)
            })
    }

    // GET: /users

    public searchUsers (callback) {
        const { accountId } = this.event

        // query dynamoDb for users via accountId
        return new Promise((resolve, reject) => {
            dynamo.getUsersByAccountId(accountId)
                .then((data) => {
                    return callback(null, { body: data.Items })
                })
                .catch(reject)
        })
    }

    public putUser (callback) {
        console.log('calling putUser')

        // Currently there are no properties in cognito that can be updated,
        // Hence we only update details in dynamoDb.
        validateUserDocument(this.event.body)
            .then(() => {
                return {
                    ...this.event.body,
                    sub: this.event.userId
                }
            })
            .then(dynamo.addUser)
            .then(() => callback(null, ResponseHandler.generateSuccessResponse(200, 'User successfully updated.')))
            .catch((err) => {
                callback(ResponseHandler.generateErrorResponseObject(err.statusCode, err))
            })
    }

    public getUser (callback) {
        const { userId } = this.event

        dynamo.getUser(userId)
            .then((data) => callback(null, { body: data }))
            .catch((err) => {
                callback(ResponseHandler.generateErrorResponseObject(err.statusCode, err.message))
            })
    }

    // Will add the patch endpoint after the first release of the API.
    // Need to make a decision about whether we use RFC 7396 or RFC 5789
    // public patchUser () {
    //     console.log('calling patchUser')
    //     console.log('this.event', this.event)
    //
    //     const { userId, body: { accountId } } = this.event
    //     dynamo.patchUser(accountId, userId, this.event.body)
    //     // Update dynamo
    //
    //     var model = {
    //         name: TABLE_NAME,   // must match DynamoDB table name
    //         partition: 'accountId', // primary partition key - cannot be reserved word (like uuid)
    //         sort: 'sub',
    //         fields: {
    //             username: "john11@citizen.com",
    //             email: "john11@citizen.com",
    //             password: "bluedot",
    //             firstName: "Bain",
    //             lastName: "McGuy",
    //             phoneNumber: "+(61) 402342234",
    //             isAdmin: false,
    //             accountId: "23a01af0-232a-4518-9c0e-323fb773f5ef",
    //             sub: userId,
    //             state: "ACTIVE"
    //         }
    //     };
    //
    //     mlFactory.create({
    //         event: this.event,
    //         context: this.context,
    //         callback: callback,
    //         model
    //     })
    //     .catch(function(err) {
    //         console.error(err)
    //         callback(err);
    //     });
    // }

    public logout (callback) {
        console.log('Authenticator: logging out...')

        const headers = this.event.headers
        const accessToken = headers.Authorization.split(' ')[1]
        const tokenType = headers.Authorization.split(' ')[0]
        let decodedJwt = jwt.decode(accessToken, { complete: true })

        if (!decodedJwt) {
            console.log('Not a valid JWT token')
            this.context.fail('Unauthorized')
            return
        }

        if (!tokenType.match(/bearer/i)) {
            console.log('Not a bearer token')
            this.context.fail('Unauthorized')
            return
        }

        const params = {
            UserPoolId: USER_POOL_ID,
            Username: decodedJwt.payload.username
        }

        this.client.adminUserGlobalSignOut(params, (err, data) => {
            if (err) {
                console.log(`Log out unsuccessful: ${decodedJwt.payload.username}`)
                console.log('Log out error', err.stack)
                callback(null, `Log out unsuccessful: ${decodedJwt.payload.username}`)
            } else {
                console.log(`logout success: ${decodedJwt.payload.username}`)
                console.log('Log out data: ', data)
                callback(null, `logout success: ${decodedJwt.payload.username}`)
            }
        })
    }

    public refresh (callback) {
        const body = this.event.body
        const headers = this.event.headers
        const refreshToken = body.refreshToken
        const accessToken = headers.Authorization.split(' ')[1]
        const decodedJwt = jwt.decode(accessToken, { complete: true })
        const username = decodedJwt.payload.username

        this.initiateRefresh(username, refreshToken)
            .then((data) => callback(null, data))
            .catch((err) => callback(err))
    }

    public deleteUser (callback) {
        console.log('authenticator: deleteUser initialising... ')

        // Get user first, to get his accountId and email for deletion
        dynamo.getUser(this.event.userId)
            .then((user) => cognito.deleteUser(user, this.client))
            .then((user) => dynamo.deleteUser(user))
            .then((user) => callback(null, ResponseHandler.generateSuccessResponse(204)))
            .catch((err) => callback(ResponseHandler.generateErrorResponseObject(404, err)))
    }

    public generatePasswordToken (callback) {
        console.log('generating password reset token... ')
        const self = this

        const params = {
            ClientId: CLIENT_ID,
            Username: this.event.body.email,
            SecretHash: this.getSecretHash(this.event.body.email)
        }

        this.client.forgotPassword(params, (err) => {
            if (err) return self.handleError(err, callback)

            // successful response
            callback(null, ResponseHandler.generateSuccessResponse(204))
        })
    }

    public resetPassword (callback) {
        const params = {
            ClientId: CLIENT_ID, /* required */
            ConfirmationCode: this.event.body.verificationCode, /* required */
            Password: this.event.body.password, /* required */
            Username: this.event.body.email, /* required */
            SecretHash: this.getSecretHash(this.event.body.email)
        }

        client.confirmForgotPassword(params, (err) => {
            if (err) return this.handleError(err, callback)

            // successful response
            callback(null, ResponseHandler.generateSuccessResponse(204))
        })
    }

    // login
    private initiateAuth (username: string, password: string, email: string, userId?: string): Promise<any> {
        return new Promise((resolve, reject) => {
            const params = {
                UserPoolId: USER_POOL_ID,
                ClientId: CLIENT_ID,
                AuthFlow: 'ADMIN_NO_SRP_AUTH',
                AuthParameters: {
                    'USERNAME': username,
                    'SECRET_HASH': this.getSecretHash(username),
                    'PASSWORD': password,
                    'EMAIL': email
                },
                ClientMetadata: {
                    'username': username,
                    'password': password,
                    'email': email
                }
            }
            console.log('initiating auth...')

            const initiateAuthPromise = this.client.adminInitiateAuth(params).promise()

            initiateAuthPromise
                .then((data: any) => {
                    AWS.config.credentials = new AWS.CognitoIdentityCredentials({
                        IdentityPoolId: IDENTITY_POOL_ID,
                        Logins: {
                            [`cognito-idp.ap-southeast-2.amazonaws.com/${USER_POOL_ID}`]: data.AuthenticationResult.IdToken
                        }
                    })

                    AWS.config.credentials['refresh']((error) => {
                        if (error) {
                            console.log('error', error)
                            return resolve(data)
                        }
                        console.log('logged into id pool')
                        data.userId = userId
                        resolve(data)
                    })
                })
                .catch((err: any) => {
                    console.log('JSON.stringify(err)', JSON.stringify(err))
                    reject(err)
                })
        })
    }

    private getSecretHash (username: string) {
        // pseudocode:
        // import javax.crypto.Mac;
        //     import javax.crypto.spec.SecretKeySpec;
        //
        // public static String calculateSecretHash(String userPoolClientId, String userPoolClientSecret, String userName) {
        //         final String HMAC_SHA256_ALGORITHM = "HmacSHA256";
        //
        //         SecretKeySpec signingKey = new SecretKeySpec(
        //             userPoolClientSecret.getBytes(StandardCharsets.UTF_8),
        //             HMAC_SHA256_ALGORITHM);
        //         try {
        //             Mac mac = Mac.getInstance(HMAC_SHA256_ALGORITHM);
        //             mac.init(signingKey);
        //             mac.update(userName.getBytes(StandardCharsets.UTF_8));
        //             byte[] rawHmac = mac.doFinal(userPoolClientId.getBytes(StandardCharsets.UTF_8));
        //             return Base64.getEncoder().encodeToString(rawHmac);
        //         } catch (Exception e) {
        //             throw new RuntimeException("Error while calculating ");
        //         }
        //     }
        const msg = username + CLIENT_ID
        const encodedMsg = CryptoJS.enc.Utf8.parse(msg)
        const encodedSecret = CryptoJS.enc.Utf8.parse(CLIENT_SECRET)
        const secretHash = CryptoJS.HmacSHA256(encodedMsg, encodedSecret)
        return secretHash.toString(CryptoJS.enc.Base64)
    }

    private handleError (err: AWS.AWSError, callback: (...args: any[]) => void) {
        // an error occurred
        console.log('err', err)
        const awsError = AWS_COGNITO_ERRORS[err.code]

        if (awsError) {
            return callback(ResponseHandler.generateErrorResponseObject(awsError.httpCode, err))
        }

        return callback(ResponseHandler.generateErrorResponseObject(400, err))

    }

    private initiateRefresh (username: string, refreshToken: string): Promise<any> {
        return new Promise((resolve, reject) => {
            const params = {
                UserPoolId: USER_POOL_ID,
                ClientId: CLIENT_ID,
                AuthFlow: 'REFRESH_TOKEN_AUTH',
                AuthParameters: {
                    'USERNAME': username,
                    'SECRET_HASH': this.getSecretHash(username),
                    'REFRESH_TOKEN': refreshToken
                },
                ClientMetadata: {
                    'username': username
                }
            }

            const initiateAuthPromise = client.adminInitiateAuth(params).promise()

            initiateAuthPromise
                .then((data: any) => {
                    resolve(data)
                })
                .catch((err: any) => {
                    console.log('err', err)
                    reject(err)
                })
        })
    }

}

export function handler (event: any , context: any, callback: (...args: any[]) => void) {
    if (!client) client = new AWS.CognitoIdentityServiceProvider({ apiVersion: '2016-04-18' })
    new Authenticator(event, context, client)[event.fun](callback)
}
