import { assert } from 'chai'
import { handler } from '../authenticator'
import { handler as authorizerHandler } from '../custom-authorizer'
const { email, postUserEvent, loginEvent, createDeleteUserEvent, createRefreshEvent, createLogoutEvent } = require('./testevents.ts')

const LambdaTester = require('lambda-tester')

describe('### Authenticator ###', () => {
    let userId
    let newLoginAccessToken
    let newLoginRefreshToken

    it('"postUser" function should create a new user then log the new user in', (done) => {
        LambdaTester(handler)
            .event(postUserEvent)
            .expectResult(({ body }) => {
                newLoginAccessToken = body.access_token
                newLoginRefreshToken = body.refresh_token
                userId = body.userId
                assert.isTrue(body.is_new)
                assert.exists(body.access_token)
                assert.exists(body.refresh_token)
                assert.exists(body.id_token)
                done()
            })
    })

    it('"login" function should log a user in', () => {
        return LambdaTester(handler)
            .event(loginEvent)
            .expectResult((result) => {
                assert.isFalse(result.is_new)
                assert.exists(result.access_token)
                assert.exists(result.refresh_token)
                assert.exists(result.id_token)
                assert.equal(result.statusCode, 201)
            })
    })

    it('"refresh" function should refresh a user access token with a valid refresh token', () => {
        const refreshEvent = createRefreshEvent(newLoginAccessToken, newLoginRefreshToken)

        return LambdaTester(handler)
            .event(refreshEvent)
            .expectResult((result) => {
                assert.exists(result.AuthenticationResult.AccessToken)
                assert.exists(result.AuthenticationResult.IdToken)
                assert.equal(
                    result.AuthenticationResult.TokenType,
                    'Bearer'
                )
            })
    })

    it('"logout" function should log a user out', () => {
        const logoutEvent = createLogoutEvent(newLoginAccessToken)

        return LambdaTester(handler)
            .event(logoutEvent)
            .expectResult((result) => {
                assert.equal(
                    result,
                    `logout success: ${email}`
                )
            })
    })

    it('should delete a user', () => {
        const deleteUserEvent = createDeleteUserEvent(userId)

        return LambdaTester(handler)
            .event(deleteUserEvent)
            .expectResult((result) => {
                assert.equal(result.statusCode, 204)
            })
    })
})

describe('### Authorizer ###', () => {
    let userId
    let newLoginAccessToken
    let newLogin = JSON.parse(JSON.stringify(loginEvent))
    newLogin.body.email = `${Math.random()
        .toString(36)
        .substr(2, 5)}@bluedotinnovation.com`

    // creates a new user and logs the user in
    before((done) => {
        LambdaTester(handler)
            .event(postUserEvent)
            .expectResult(({ body }) => {
                userId = body.userId
                newLoginAccessToken = body.access_token
                assert.isTrue(body.is_new)
                assert.exists(body.access_token)
                assert.exists(body.refresh_token)
                assert.exists(body.id_token)
                done()
            })
    })

    after(() => {
        const deleteUserEvent = createDeleteUserEvent(userId)

        return LambdaTester(handler)
            .event(deleteUserEvent)
            .expectResult((result) => {
                assert.equal(result.statusCode, 204)
            })
    })

    it('Should generate an IAM policy', () => {
        function cb () {
            console.log('bla')
        }
        let auth = {
            type: 'TOKEN',
            methodArn:
                'arn:aws:execute-api:ap-southeast-2:791031460737:ku81jipf51/v2c/POST/zones',
            authorizationToken: `Bearer ${newLoginAccessToken}`
        }

        authorizerHandler(
            auth,
            {
                succeed: (data) => {
                    assert.exists(data.principalId)
                    assert.exists(data.policyDocument)
                },
                fail: (err) => {
                    console.log('err', err)
                }
            },
            cb
        )
    })
})
